"""
Module for neural models to use during experiment.

To work properly, each customised model needs to inherit from AbstractModelClass and needs an implementation of the
set_model method.

In this module, you can find some exemplary model classes that have been build and were running in a experiment.

* `MyLittleModel`: small model implementation with a single 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time).
* `MyBranchedModel`: a model with single  1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), it has three
  output branches from different layers of the model.
* `MyTowerModel`: a more complex model with inception blocks (called towers)
* `MyPaperModel`: A model used for the publication: <Add Publication Title / Citation>

In addition, a short introduction how to create your own model is given hereinafter.

How to create a customised model?
#################################

* Create a new class:

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def __init__(self, input_shape: list, output_shape: list):

                super().__init__(input_shape[0], output_shape[0])

                # settings
                self.dropout_rate = 0.1
                self.activation = keras.layers.PReLU

                # apply to model
                self.set_model()
                self.set_compile_options()
                self.set_custom_objects(loss=self.compile_options['loss'])

* Make sure to add the `super().__init__()` and at least `set_model()` and `set_compile_options()` to your custom init
  method.
* If you have custom objects in your model, that are not part of keras, you need to add them to custom objects. To do
  this, call `set_custom_objects` with arbitrarily kwargs. In the shown example, the loss has been added, because it
  wasn't a standard loss. Apart from this, we always encourage you to add the loss as custom object, to prevent
  potential errors when loading an already created model instead of training a new one.
* Build your model inside `set_model()`, e.g.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_model(self):
                x_input = keras.layers.Input(shape=self._input_shape)
                x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1'.format("major"))(x_input)
                x_in = self.activation(name='{}_conv_act'.format("major"))(x_in)
                x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
                x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
                x_in = keras.layers.Dense(16, name='{}_Dense_16'.format("major"))(x_in)
                x_in = self.activation()(x_in)
                x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
                out_main = self.activation()(x_in)
                self.model = keras.Model(inputs=x_input, outputs=[out_main])

* Your are free, how to design your model. Just make sure to save it in the class attribute model.
* Additionally, set your custom compile options including the loss.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_compile_options(self):
                self.initial_lr = 1e-2
                self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
                self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr,
                                                                                       drop=.94,
                                                                                       epochs_drop=10)
                self.loss = keras.losses.mean_squared_error
                self.compile_options = {"metrics": ["mse", "mae"]}

* If you have a branched model with multiple outputs, you need either set only a single loss for all branch outputs or
  to provide the same number of loss functions considering the right order. E.g.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_model(self):
                ...
                self.model = keras.Model(inputs=x_input, outputs=[out_minor_1, out_minor_2, out_main])

            def set_compile_options(self):
                self.loss = [keras.losses.mean_absolute_error] +  # for out_minor_1
                            [keras.losses.mean_squared_error] +   # for out_minor_2
                            [keras.losses.mean_squared_error]     # for out_main


How to access my customised model?
##################################

If the customised model is created, you can easily access the model with

>>> MyCustomisedModel().model
<your custom model>

The loss is accessible via

>>> MyCustomisedModel().loss
<your custom loss>

You can treat the instance of your model as instance but also as the model itself. If you call a method, that refers to
the model instead of the model instance, you can directly apply the command on the instance instead of adding the model
parameter call.

>>> MyCustomisedModel().model.compile(**kwargs) == MyCustomisedModel().compile(**kwargs)
True

"""

import mlair.model_modules.keras_extensions

__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2020-05-12'

from abc import ABC
from typing import Any, Callable, Dict

import keras
import tensorflow as tf
from mlair.model_modules.inception_model import InceptionModelBase
from mlair.model_modules.flatten import flatten_tail
from mlair.model_modules.advanced_paddings import PadUtils, Padding2D


class AbstractModelClass(ABC):
    """
    The AbstractModelClass provides a unified skeleton for any model provided to the machine learning workflow.

    The model can always be accessed by calling ModelClass.model or directly by an model method without parsing the
    model attribute name (e.g. ModelClass.model.compile -> ModelClass.compile). Beside the model, this class provides
    the corresponding loss function.
    """

    def __init__(self, input_shape, output_shape) -> None:
        """Predefine internal attributes for model and loss."""
        self.__model = None
        self.model_name = self.__class__.__name__
        self.__custom_objects = {}
        self.__allowed_compile_options = {'optimizer': None,
                                          'loss': None,
                                          'metrics': None,
                                          'loss_weights': None,
                                          'sample_weight_mode': None,
                                          'weighted_metrics': None,
                                          'target_tensors': None
                                          }
        self.__compile_options = self.__allowed_compile_options
        self.__compile_options_is_set = False
        self._input_shape = input_shape
        self._output_shape = self.__extract_from_tuple(output_shape)

    def __getattr__(self, name: str) -> Any:
        """
        Is called if __getattribute__ is not able to find requested attribute.

        Normally, the model class is saved into a variable like `model = ModelClass()`. To bypass a call like
        `model.model` to access the _model attribute, this method tries to search for the named attribute in the
        self.model namespace and returns this attribute if available. Therefore, following expression is true:
        `ModelClass().compile == ModelClass().model.compile` as long the called attribute/method is not part if the
        ModelClass itself.

        :param name: name of the attribute or method to call

        :return: attribute or method from self.model namespace
        """
        return self.model.__getattribute__(name)

    @property
    def model(self) -> keras.Model:
        """
        The model property containing a keras.Model instance.

        :return: the keras model
        """
        return self.__model

    @model.setter
    def model(self, value):
        self.__model = value

    @property
    def custom_objects(self) -> Dict:
        """
        The custom objects property collects all non-keras utilities that are used in the model class.

        To load such a customised and already compiled model (e.g. from local disk), this information is required.

        :return: custom objects in a dictionary
        """
        return self.__custom_objects

    @custom_objects.setter
    def custom_objects(self, value) -> None:
        self.__custom_objects = value

    @property
    def compile_options(self) -> Callable:
        """
        The compile options property allows the user to use all keras.compile() arguments. They can ether be passed as
        dictionary (1), as attribute, without setting compile_options (2) or as mixture (partly defined as instance
        attributes and partly parsing a dictionary) of both of them (3).
        The method will raise an Error when the same parameter is set differently.

        Example (1) Recommended (includes check for valid keywords which are used as args in keras.compile)
        .. code-block:: python
            def set_compile_options(self):
                self.compile_options = {"optimizer": keras.optimizers.SGD(),
                                        "loss": keras.losses.mean_squared_error,
                                        "metrics": ["mse", "mae"]}

        Example (2)
        .. code-block:: python
            def set_compile_options(self):
                self.optimizer = keras.optimizers.SGD()
                self.loss = keras.losses.mean_squared_error
                self.metrics = ["mse", "mae"]

        Example (3)
        Correct:
        .. code-block:: python
            def set_compile_options(self):
                self.optimizer = keras.optimizers.SGD()
                self.loss = keras.losses.mean_squared_error
                self.compile_options = {"metrics": ["mse", "mae"]}

        Incorrect: (Will raise an error)
        .. code-block:: python
            def set_compile_options(self):
                self.optimizer = keras.optimizers.SGD()
                self.loss = keras.losses.mean_squared_error
                self.compile_options = {"optimizer" = keras.optimizers.Adam(), "metrics": ["mse", "mae"]}

        Note:
        * As long as the attribute and the dict value have exactly the same values, the setter method will not raise
        an error
        * For example (2) there is no check implemented, if the attributes are valid compile options


        :return:
        """
        if self.__compile_options_is_set is False:
            self.compile_options = None
        return self.__compile_options

    @compile_options.setter
    def compile_options(self, value: Dict) -> None:
        if isinstance(value, dict):
            if not (set(value.keys()) <= set(self.__allowed_compile_options.keys())):
                raise ValueError(f"Got invalid key for compile_options. {value.keys()}")

        for allow_k in self.__allowed_compile_options.keys():
            if hasattr(self, allow_k):
                new_v_attr = getattr(self, allow_k)
            else:
                new_v_attr = None
            if isinstance(value, dict):
                new_v_dic = value.pop(allow_k, None)
            elif value is None:
                new_v_dic = None
            else:
                raise TypeError(f"`compile_options' must be `dict' or `None', but is {type(value)}.")
            if (new_v_attr == new_v_dic or self.__compare_keras_optimizers(new_v_attr, new_v_dic)) or (
                    (new_v_attr is None) ^ (new_v_dic is None)):
                if new_v_attr is not None:
                    self.__compile_options[allow_k] = new_v_attr
                else:
                    self.__compile_options[allow_k] = new_v_dic

            else:
                raise ValueError(
                    f"Got different values or arguments for same argument: self.{allow_k}={new_v_attr.__class__} and '{allow_k}': {new_v_dic.__class__}")
        self.__compile_options_is_set = True

    @staticmethod
    def __extract_from_tuple(tup):
        """Return element of tuple if it contains only a single element."""
        return tup[0] if isinstance(tup, tuple) and len(tup) == 1 else tup

    @staticmethod
    def __compare_keras_optimizers(first, second):
        """
        Compares if optimiser and all settings of the optimisers are exactly equal.

        :return True if optimisers are interchangeable, or False if optimisers are distinguishable.
        """
        if first.__class__ == second.__class__ and first.__module__ == 'keras.optimizers':
            res = True
            init = tf.global_variables_initializer()
            with tf.Session() as sess:
                sess.run(init)
                for k, v in first.__dict__.items():
                    try:
                        res *= sess.run(v) == sess.run(second.__dict__[k])
                    except TypeError:
                        res *= v == second.__dict__[k]
        else:
            res = False
        return bool(res)

    def get_settings(self) -> Dict:
        """
        Get all class attributes that are not protected in the AbstractModelClass as dictionary.

        :return: all class attributes
        """
        return dict((k, v) for (k, v) in self.__dict__.items() if not k.startswith("_AbstractModelClass__"))

    def set_model(self):
        """Abstract method to set model."""
        raise NotImplementedError

    def set_compile_options(self):
        """
        This method only has to be defined in child class, when additional compile options should be used ()
        (other options than optimizer and loss)
        Has to be set as dictionary: {'optimizer': None,
                                      'loss': None,
                                      'metrics': None,
                                      'loss_weights': None,
                                      'sample_weight_mode': None,
                                      'weighted_metrics': None,
                                      'target_tensors': None
                                      }

        :return:
        """
        raise NotImplementedError

    def set_custom_objects(self, **kwargs) -> None:
        """
        Set custom objects that are not part of keras framework.

        These custom objects are needed if an already compiled model is loaded from disk. There is a special treatment
        for the Padding2D class, which is a base class for different padding types. For a correct behaviour, all
        supported subclasses are added as custom objects in addition to the given ones.

        :param kwargs: all custom objects, that should be saved
        """
        if "Padding2D" in kwargs.keys():
            kwargs.update(kwargs["Padding2D"].allowed_paddings)
        self.custom_objects = kwargs


class MyLittleModel(AbstractModelClass):
    """
    A customised model 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the output layer depending
    on the window_lead_time parameter.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.1
        self.regularizer = keras.regularizers.l2(0.1)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        """
        Build the model.
        """
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Flatten(name='{}'.format("major"))(x_input)
        x_in = keras.layers.Dense(64, name='{}_Dense_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(32, name='{}_Dense_32'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(16, name='{}_Dense_16'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
        out_main = self.activation()(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-2
        self.optimizer = keras.optimizers.adam(lr=self.initial_lr)
        # self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
        #                                                                        epochs_drop=10)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class MyLittleModelHourly(AbstractModelClass):
    """
    A customised model with a 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the
    output layer depending on the window_lead_time parameter. Dropout is used between the Convolution and the first
    Dense layer.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param shape_inputs: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param shape_outputs: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.1
        self.regularizer = keras.regularizers.l2(0.001)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        """
        Build the model.
        """

        # add 1 to window_size to include current time step t0
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(128, (1, 1), padding='same', name='{}_Conv_1x1_128'.format("major"))(x_input)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(64, (1, 1), padding='same', name='{}_Conv_1x1_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1_32'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
        x_in = keras.layers.Dense(128, name='{}_Dense_128'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(64, name='{}_Dense_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
        out_main = self.activation()(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-2
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class MyBranchedModel(AbstractModelClass):
    """
    A customised model

    with a 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the
    output layer depending on the window_lead_time parameter. Dropout is used between the Convolution and the first
    Dense layer.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.1
        self.regularizer = keras.regularizers.l2(0.1)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"])

    def set_model(self):
        """
        Build the model.
        """

        # add 1 to window_size to include current time step t0
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1'.format("major"))(x_input)
        x_in = self.activation(name='{}_conv_act'.format("major"))(x_in)
        x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
        x_in = keras.layers.Dense(64, name='{}_Dense_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        out_minor_1 = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("minor_1"))(x_in)
        out_minor_1 = self.activation(name="minor_1")(out_minor_1)
        x_in = keras.layers.Dense(32, name='{}_Dense_32'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        out_minor_2 = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("minor_2"))(x_in)
        out_minor_2 = self.activation(name="minor_2")(out_minor_2)
        x_in = keras.layers.Dense(16, name='{}_Dense_16'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
        out_main = self.activation(name="main")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_minor_1, out_minor_2, out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-2
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
                                                                               epochs_drop=10)
        self.compile_options = {"loss": [keras.losses.mean_absolute_error] + [keras.losses.mean_squared_error] + [
            keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class MyTowerModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 1e-2
        self.regularizer = keras.regularizers.l2(0.1)
        self.initial_lr = 1e-2
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
                                                                               epochs_drop=10)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"])

    def set_model(self):
        """
        Build the model.
        """
        activation = self.activation
        conv_settings_dict1 = {
            'tower_1': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (3, 1), 'activation': activation},
            'tower_2': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (5, 1), 'activation': activation},
            'tower_3': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (1, 1), 'activation': activation},
        }

        pool_settings_dict1 = {'pool_kernel': (3, 1), 'tower_filter': 8 * 2, 'activation': activation}

        conv_settings_dict2 = {
            'tower_1': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (1, 1),
                        'activation': activation},
        }
        pool_settings_dict2 = {'pool_kernel': (3, 1), 'tower_filter': 16, 'activation': activation}

        conv_settings_dict3 = {'tower_1': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (3, 1),
                                           'activation': activation},
                               'tower_2': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (5, 1),
                                           'activation': activation},
                               'tower_3': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (1, 1),
                                           'activation': activation},
                               }

        pool_settings_dict3 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        ##########################################
        inception_model = InceptionModelBase()

        X_input = keras.layers.Input(shape=self._input_shape)

        X_in = inception_model.inception_block(X_input, conv_settings_dict1, pool_settings_dict1,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict2, pool_settings_dict2,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict3, pool_settings_dict3,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)
        #############################################

        # out_main = flatten_tail(X_in, 'Main', activation=activation, bound_weight=True, dropout_rate=self.dropout_rate,
        #                         reduction_filter=64, inner_neurons=64, output_neurons=self.window_lead_time)

        out_main = flatten_tail(X_in, inner_neurons=64, activation=activation, output_neurons=self._output_shape,
                                output_activation='linear', reduction_filter=64,
                                name='Main', bound_weight=True, dropout_rate=self.dropout_rate,
                                kernel_regularizer=self.regularizer
                                )

        self.model = keras.Model(inputs=X_input, outputs=[out_main])

    def set_compile_options(self):
        self.optimizer = keras.optimizers.adam(lr=self.initial_lr)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse"]}


class MyPaperModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = .3
        self.regularizer = keras.regularizers.l2(0.001)
        self.initial_lr = 1e-3
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
                                                                               epochs_drop=10)
        self.activation = keras.layers.ELU
        self.padding = "SymPad2D"

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"], Padding2D=Padding2D)

    def set_model(self):
        """
        Build the model.

        :param activation: activation function
        :param window_history_size: number of historical time steps included in the input data
        :param channels: number of variables used in input data
        :param dropout_rate: dropout rate used in the model [0, 1)
        :param window_lead_time: number of time steps to forecast in the output layer
        :return: built keras model
        """
        activation = self.activation
        first_kernel = (3, 1)
        first_filters = 16

        conv_settings_dict1 = {
            'tower_1': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (1, 1),
                        'activation': activation},
            # 'tower_4':{'reduction_filter':8, 'tower_filter':8*2, 'tower_kernel':(7,1), 'activation':activation},
        }
        pool_settings_dict1 = {'pool_kernel': (3, 1), 'tower_filter': 16, 'activation': activation}

        conv_settings_dict2 = {
            'tower_1': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (1, 1),
                        'activation': activation},
            # 'tower_4':{'reduction_filter':8*2, 'tower_filter':16*2, 'tower_kernel':(7,1), 'activation':activation},
        }
        pool_settings_dict2 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        conv_settings_dict3 = {
            'tower_1': {'reduction_filter': 64 * 2, 'tower_filter': 32 * 4, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 64 * 2, 'tower_filter': 32 * 4, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 64 * 2, 'tower_filter': 32 * 4, 'tower_kernel': (1, 1),
                        'activation': activation},
            # 'tower_4':{'reduction_filter':16*4, 'tower_filter':32, 'tower_kernel':(7,1), 'activation':activation},
        }
        pool_settings_dict3 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        ##########################################
        inception_model = InceptionModelBase()

        X_input = keras.layers.Input(shape=self._input_shape)

        pad_size = PadUtils.get_padding_for_same(first_kernel)
        # X_in = adv_pad.SymmetricPadding2D(padding=pad_size)(X_input)
        # X_in = inception_model.padding_layer("SymPad2D")(padding=pad_size, name="SymPad")(X_input)  # adv_pad.SymmetricPadding2D(padding=pad_size)(X_input)
        X_in = Padding2D("SymPad2D")(padding=pad_size, name="SymPad")(X_input)
        X_in = keras.layers.Conv2D(filters=first_filters,
                                   kernel_size=first_kernel,
                                   kernel_regularizer=self.regularizer,
                                   name="First_conv_{}x{}".format(first_kernel[0], first_kernel[1]))(X_in)
        X_in = self.activation(name='FirstAct')(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict1, pool_settings_dict1,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True,
                                               padding=self.padding)
        # out_minor1 = flatten_tail(X_in, 'minor_1', False, self.dropout_rate, self.window_lead_time,
        #                           self.activation, 32, 64)
        out_minor1 = flatten_tail(X_in, inner_neurons=64, activation=activation, output_neurons=self._output_shape,
                                  output_activation='linear', reduction_filter=32,
                                  name='minor_1', bound_weight=False, dropout_rate=self.dropout_rate,
                                  kernel_regularizer=self.regularizer
                                  )

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict2, pool_settings_dict2,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True, padding=self.padding)

        # X_in = keras.layers.Dropout(self.dropout_rate)(X_in)
        #
        # X_in = inception_model.inception_block(X_in, conv_settings_dict3, pool_settings_dict3, regularizer=self.regularizer,
        #                                        batch_normalisation=True)
        #############################################

        out_main = flatten_tail(X_in, inner_neurons=64 * 2, activation=activation, output_neurons=self._output_shape,
                                output_activation='linear', reduction_filter=64 * 2,
                                name='Main', bound_weight=False, dropout_rate=self.dropout_rate,
                                kernel_regularizer=self.regularizer
                                )

        self.model = keras.Model(inputs=X_input, outputs=[out_minor1, out_main])

    def set_compile_options(self):
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.compile_options = {"loss": [keras.losses.mean_squared_error, keras.losses.mean_squared_error],
                                "metrics": ['mse', 'mea']}


if __name__ == "__main__":
    model = MyLittleModel([(1, 3, 10)], [2])
    print(model.compile_options)
